@SuppressWarnings("SuspiciousNameCombination")
public class AVLDrevo {
    static class Node {
        int value;
        int count = 1;
        Node left, right;
        int height = 1;

        public Node(int value) {
            this.value = value;
        }

        @Override
        public String toString() {
            return value + "/" + count;
        }
    }

    Node root;
    private StringBuilder traversal = new StringBuilder();

    private int getHeight(Node node) {
        if (node == null) {
            return 0;
        }
        return node.height;
    }

    private int getBalance(Node node) {
        if (node == null) {
            return 0;
        }
        return getHeight(node.left) - getHeight(node.right);
    }

    private Node rightRotate(Node y) {
        Node x = y.left;
        y.left = x.right;
        x.right = y;
        y.height = 1 + Math.max(getHeight(y.left), getHeight(y.right));
        x.height = 1 + Math.max(getHeight(x.left), getHeight(x.right));
        return x;
    }

    private Node leftRotate(Node x) {
        Node y = x.right;
        x.right = y.left;
        y.left = x;
        x.height = 1 + Math.max(getHeight(x.left), getHeight(x.right));
        y.height = 1 + Math.max(getHeight(y.left), getHeight(y.right));
        return y;
    }

    private Node maxValueNode(Node node) {
        Node current = node;
        while (current.right != null) {
            current = current.right;
        }
        return current;
    }

    private Node balance(Node node) {
        int balance = getBalance(node);

        if (balance > 1 && getBalance(node.left) >= 0) {
            return rightRotate(node);
        }

        if (balance > 1 && getBalance(node.left) < 0) {
            node.left = leftRotate(node.left);
            return rightRotate(node);
        }

        if (balance < -1 && getBalance(node.right) <= 0) {
            return leftRotate(node);
        }

        if (balance < -1 && getBalance(node.right) > 0) {
            node.right = rightRotate(node.right);
            return leftRotate(node);
        }

        return node;
    }

    public void vstavi(int kljuc) {
        root = vstavi(root, kljuc);
    }

    private Node vstavi(Node node, int value) {
        if (node == null) {
            return new Node(value);
        }

        if (value == node.value) {
            node.count++;
            return node;
        } else if (value < node.value) {
            node.left = vstavi(node.left, value);
        } else {
            node.right = vstavi(node.right, value);
        }

        node.height = 1 + Math.max(getHeight(node.left), getHeight(node.right));

        return balance(node);
    }

    public void najdi(int kljuc) {
        traversal = new StringBuilder();
        najdi(root, kljuc);
        System.out.println(traversal.substring(0, traversal.length() - 1));
    }

    private Node najdi(Node node, int value) {
        if (node == null) {
            traversal.append("x").append(",");
            return null;
        }

        if (value == node.value) {
            traversal.append(node.value).append(",");
            return node;
        } else if (value < node.value) {
            traversal.append(node.value).append(",");
            return najdi(node.left, value);
        } else {
            traversal.append(node.value).append(",");
            return najdi(node.right, value);
        }
    }

    public void izbrisi(int kljuc) {
        root = izbrisi(root, kljuc);
    }

    private Node izbrisi(Node node, int value) {
        if (node == null) {
            return null;
        }

        if (value < node.value) {
            node.left = izbrisi(node.left, value);
        } else if (value > node.value) {
            node.right = izbrisi(node.right, value);
        } else {
            if (node.count > 1) {
                node.count--;
                return node;
            }

            if (node.left == null) {
                return node.right;
            } else if (node.right == null) {
                return node.left;
            }

            Node max = maxValueNode(node.left);
            node.value = max.value;
            node.count = max.count;
            max.count = 1;
            node.left = izbrisi(node.left, max.value);
        }

        node.height = 1 + Math.max(getHeight(node.left), getHeight(node.right));

        return balance(node);
    }

    public void premiPregled() {
        traversal = new StringBuilder();
        if (root == null) {
            System.out.println("empty");
            return;
        }
        premiPregled(root);
        System.out.println(traversal.substring(0, traversal.length() - 1));
    }

    private void premiPregled(Node node) {
        if (node == null) {
            return;
        }
        traversal.append(node).append(",");
        premiPregled(node.left);
        premiPregled(node.right);
    }

    public void najnizjiSkupniPredhodnik(int a, int b) {
        if (najdi(root, a) == null || najdi(root, b) == null) {
            System.out.println("x");
            return;
        }

        Node node = lca(root, a, b);
        if (node == null) {
            System.out.println("x");
            return;
        }
        System.out.println(node.value);
    }

    private Node lca(Node node, int a, int b) {
        if (node == null) {
            return null;
        }

        if (a < node.value && b < node.value) {
            return lca(node.left, a, b);
        } else if (a > node.value && b > node.value) {
            return lca(node.right, a, b);
        }

        return node;
    }

    public void vsotaVMejah(int spodnjaMeja, int zgornjaMeja) {
        int sum = sumBounds(root, spodnjaMeja, zgornjaMeja);
        System.out.println(sum);
    }

    private int sumBounds(Node node, int lower, int upper) {
        if (node == null) {
            return 0;
        }

        int sum = 0;
        if (node.value >= lower && node.value <= upper) {
            sum += node.value * node.count;
        }

        sum += sumBounds(node.left, lower, upper);
        sum += sumBounds(node.right, lower, upper);

        return sum;
    }

    public void ktiNajmanjsi(int indeks) {
        Integer smallest = kSmallest(root, indeks);
        if (smallest != null) {
            System.out.println("x");
        }
    }

    private Integer kSmallest(Node node, int index) {
        if (node == null) {
            return index;
        }
        Integer k = kSmallest(node.left, index);
        if (k == null) {
            return null;
        }
        k -= node.count;
        if (k <= 0) {
            System.out.println(node.value);
            return null;
        }
        return kSmallest(node.right, k);
    }

    static void beautyPrint(AVLDrevo.Node node, int l) {
        if (node == null)
            return;

        for (int i = 0; i < l; i++) {
            System.out.print("\t");
        }
        System.out.println(node);
        beautyPrint(node.left, l + 1);
        beautyPrint(node.right, l + 1);
    }
}
