public class Graf {
    private final MaxHeap<Node>.HeapNode[] nodes;
    private final MaxHeap<Node> heap = new MaxHeap<>();

    @SuppressWarnings("unchecked")
    Graf(int n) {
        this.nodes = new MaxHeap.HeapNode[n];
        for (int i = 0; i < n; i++) {
            Node node = new Node(i);
            nodes[i] = heap.insert(node);
        }
    }

    void povezi(int u, int v) {
        nodes[u].value.degree++;
        nodes[u].value.neighbors.add(nodes[v]);
        heap.siftUp(nodes[u].index);

        nodes[v].value.degree++;
        nodes[v].value.neighbors.add(nodes[u]);
        heap.siftUp(nodes[v].index);
    }

    void grobaSila() {

    }

    void pozresni() {
        MaxHeap<Node> heap = new MaxHeap<>(this.heap);
        for (var heapNode : nodes) {
            var node = heapNode.value;
            node.degree = node.neighbors.size();
        }

        ArrList<Integer> used = new ArrList<>();

        while (!heap.isEmpty()) {
            Node node = heap.pop();
            if (node.degree == 0) {
                break;
            }

            for (int i = 0; i < node.neighbors.size(); i++) {
                var neighbor = node.neighbors.get(i);
                if (neighbor.value.degree > 0) {
                    neighbor.value.degree--;
                    heap.siftDown(neighbor.index);
                }
            }

            used.add(node.id);
        }

        print(used);
    }

    private void print(ArrList<Integer> arr) {
        int idxStart = arr.size() > 100 ? arr.size() - 100 : 0;
        for (int i = idxStart; i < arr.size(); i++) {
            System.out.print(arr.get(i));
            if (i < arr.size() - 1) System.out.print(" ");
        }
        System.out.println();
    }
}

class Node implements Comparable<Node> {
    int id;
    int degree = 0;
    //    LinkedList<MaxHeap<Node>.HeapNode> neighbors = new LinkedList<>();
    ArrList<MaxHeap<Node>.HeapNode> neighbors = new ArrList<>();

    Node(int id) {
        this.id = id;
    }

    @Override
    public int compareTo(Node other) {
        int deg = Integer.compare(this.degree, other.degree);
        if (deg == 0) {
            return Integer.compare(other.id, this.id);
        }
        return deg;
    }

    @Override
    public String toString() {
        return "Node [id=" + id + ", degree=" + degree + "]";
    }
}

class MaxHeap<T extends Comparable<T>> {
    class HeapNode implements Comparable<HeapNode> {
        int index;
        T value;

        HeapNode(T value, int index) {
            this.value = value;
            this.index = index;
        }

        @Override
        public int compareTo(HeapNode o) {
            return this.value.compareTo(o.value);
        }

        @Override
        public String toString() {
            return this.value.toString();
        }
    }

    final private ArrList<HeapNode> arr;

    public MaxHeap() {
        arr = new ArrList<>();
    }

    public MaxHeap(MaxHeap<T> heap) {
        this.arr = new ArrList<>(heap.arr);
    }

    private int parent(int i) {
        return (i - 1) / 2;
    }

    private int leftChild(int i) {
        return 2 * i + 1;
    }

    private int rightChild(int i) {
        return 2 * i + 2;
    }

    private void swap(int i, int j) {
        HeapNode nodeI = arr.get(i);
        HeapNode nodeJ = arr.get(j);
        nodeI.index = j;
        nodeJ.index = i;
        arr.set(i, nodeJ);
        arr.set(j, nodeI);
    }

    public void siftDown(int i) {
        int largest = i;
        int left = leftChild(i);
        int right = rightChild(i);

        if (left < arr.size() && arr.get(left).compareTo(arr.get(largest)) > 0) {
            largest = left;
        }

        if (right < arr.size() && arr.get(right).compareTo(arr.get(largest)) > 0) {
            largest = right;
        }

        if (largest != i) {
            swap(i, largest);
            siftDown(largest);
        }
    }

    public void siftUp(int i) {
        int parent = parent(i);

        if (parent < 0) return;

        if (arr.get(i).compareTo(arr.get(parent)) > 0) {
            swap(i, parent);
            siftUp(parent);
        }
    }

    public HeapNode insert(T x) {
        HeapNode node = new HeapNode(x, arr.size());
        arr.add(node);
        int currentIndex = arr.size() - 1;
        while (currentIndex > 0 && arr.get(currentIndex).compareTo(arr.get(parent(currentIndex))) > 0) {
            swap(currentIndex, parent(currentIndex));
            currentIndex = parent(currentIndex);
        }
        return node;
    }

    public boolean isEmpty() {
        return arr.isEmpty();
    }

    public void rebuild() {
        for (int i = parent(arr.size() - 1); i >= 0; i--) {
            siftDown(i);
        }
    }

    public T pop() {
        if (arr.isEmpty()) {
            throw new RuntimeException("Heap is empty");
        }

        HeapNode max = arr.get(0);
        HeapNode lastElement = arr.removeLast();

        if (!arr.isEmpty()) {
            arr.set(0, lastElement);
            lastElement.index = 0;
            siftDown(0);
        }

        return max.value;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < arr.size(); i++) {
            sb.append(arr.get(i)).append(" ");
        }
        return sb.toString();
    }
}

class ArrList<T> {
    private T[] array;
    private int size;

    @SuppressWarnings("unchecked")
    public ArrList() {
        array = (T[]) new Object[64];
        size = 0;
    }

    @SuppressWarnings("unchecked")
    public ArrList(ArrList<T> list) {
        this.size = list.size;
        this.array = (T[]) new Object[list.array.length];
        System.arraycopy(list.array, 0, this.array, 0, list.size);
    }

    private boolean isFull() {
        return size == array.length;
    }

    private void resize() {
        @SuppressWarnings("unchecked")
        T[] tmp = (T[]) new Object[array.length * 2];
        System.arraycopy(array, 0, tmp, 0, size);
        array = tmp;
    }

    public T get(int i) {
        return array[i];
    }

    public void add(T t) {
        if (isFull()) {
            resize();
        }
        array[size++] = t;
    }

    public void set(int i, T t) {
        array[i] = t;
    }

    public T removeLast() {
        T tmp = array[--size];
        array[size] = null;
        return tmp;
    }

    public int size() {
        return size;
    }

    public boolean isEmpty() {
        return size == 0;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < size; i++) {
            sb.append(array[i]).append(" ");
        }
        return sb.toString();
    }
}
